/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#define  __NO_VERSION__

#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"

#if defined(NV_LINUX_ACPI_EVENTS_SUPPORTED)

/*
 * This code allows RM to subscribe to events surrounding the power adapter.
 *
 */

/*
 * c file definitions of function that do not need external visibility
 *
 */
static int         nv_acpi_add         (struct acpi_device *device);
static int         nv_acpi_remove      (struct acpi_device *device, int type);
static void        nv_acpi_event       (acpi_handle, u32, void *);
static acpi_status nv_acpi_find_nvif   (acpi_handle, u32, void *, void **);

#if defined(NV_ACPI_DEVICE_OPS_HAS_MATCH)
static int         nv_acpi_match       (struct acpi_device *, struct acpi_driver *);
#endif

#ifdef DEBUG
/*
 * internal debug routines, very useful for viewing the namespace of the VGA device
 */
static void        nv_acpi_dump_device (acpi_handle handle);
static acpi_status nv_acpi_walk_callback (acpi_handle, u32, void *, void **);
#endif //DEBUG

#if defined(ACPI_VIDEO_HID) && defined(NV_ACPI_DEVICE_ID_HAS_DRIVER_DATA) 
static const struct acpi_device_id nv_video_device_ids[] = {
    { 
        .id          = ACPI_VIDEO_HID, 
        .driver_data = 0, 
    },
    { 
        .id          = "",
        .driver_data = 0, 
    },
};
#endif

static struct acpi_driver *nv_acpi_driver;
static acpi_handle nvif_handle = NULL;

static const struct acpi_driver nv_acpi_driver_template = {
    .name = "NVIDIA ACPI Video Driver",
    .class = "video",
#if defined(ACPI_VIDEO_HID)
#if defined(NV_ACPI_DEVICE_ID_HAS_DRIVER_DATA)
    .ids = nv_video_device_ids,
#else
    .ids = ACPI_VIDEO_HID,
#endif
#endif
    .ops = {
        .add = nv_acpi_add,
        .remove = nv_acpi_remove,
#if defined(NV_ACPI_DEVICE_OPS_HAS_MATCH)
        .match = nv_acpi_match,
#endif
    },
};

int nv_acpi_init(void)
{
    /*
     * This function will register the RM with the Linux
     * ACPI subsystem.
     */
    int status;

    if (nv_acpi_driver != NULL)
        return -EBUSY;

    os_alloc_mem((void **)&nv_acpi_driver, sizeof(struct acpi_driver));
    if (nv_acpi_driver == NULL)
        return -ENOMEM;

    memcpy((void *)nv_acpi_driver, (void *)&nv_acpi_driver_template,
            sizeof(struct acpi_driver));

    status = acpi_bus_register_driver(nv_acpi_driver);
    if (status < 0)
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_init: acpi_bus_register_driver() failed (%d)!\n", status);
        os_free_mem(nv_acpi_driver);
        nv_acpi_driver = NULL;
    }

    return status;
}

int nv_acpi_uninit(void)
{
    if (nv_acpi_driver == NULL)
        return -ENXIO;

    acpi_bus_unregister_driver(nv_acpi_driver);
    os_free_mem(nv_acpi_driver);

    nv_acpi_driver = NULL;

    return 0;
}

static int nv_acpi_add(struct acpi_device *device)
{
    /*
     * This function will cause RM to initialize the things it needs for acpi interaction
     * on the display device.
     */
    int status = -1;
    RM_STATUS rmStatus = RM_ERROR;
    nv_acpi_t *pNvAcpiObject = NULL;
    union acpi_object control_argument_0 = { ACPI_TYPE_INTEGER };
    struct acpi_object_list control_argument_list = { 0, NULL };
    nv_stack_t *sp = NULL;
    struct list_head *node, *next;
    unsigned long device_id = 0;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate stack!\n");
        return -ENOMEM;
    }

    // allocate data structure we need
    rmStatus = os_alloc_mem((void **) &pNvAcpiObject, sizeof(nv_acpi_t));
    if (rmStatus != RM_OK)
    {
        NV_KMEM_CACHE_FREE_STACK(sp);
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_add: failed to allocate ACPI device management data!\n");
        return -ENOMEM;
    }

    os_mem_set((void *)pNvAcpiObject, 0, sizeof(nv_acpi_t));

    // assign driver data structure ptr to this device
    acpi_driver_data(device) = pNvAcpiObject;

    // store a device reference in our object
    pNvAcpiObject->device = device;
    pNvAcpiObject->sp = sp;

    // grab handles to all the important nodes representing devices

    list_for_each_safe(node, next, &device->children) 
    {
        struct acpi_device *dev =
            list_entry(node, struct acpi_device, node);

        if (!dev)
            continue;

        status =
            acpi_evaluate_integer(dev->handle, "_ADR", NULL, &device_id);
        if (ACPI_FAILURE(status))
            /* Couldnt query device_id for this device */
            continue;

        device_id = (device_id & 0xffff);

        /* Check if the device_id is amongst the set of 
           possible device-ids for a CRT ...*/
        if ((device_id >= 0x0100) && (device_id <= 0x0107))
        {
            pNvAcpiObject->crt_handle = dev->handle;
        }
        /* the set of possible device-ids for a TV */
        else if ((device_id >= 0x0200) && (device_id <= 0x0207))
        {
            pNvAcpiObject->tv_handle = dev->handle;
        }
        else if ((device_id == 0x0110) ||  /* device id for internal LCD */
                 (device_id == 0x0118) ||  /* alternate ACPI ID for the 
                                                            internal LCD */
                 (device_id == 0x0400))    /* ACPI spec 3.0 specified 
                                             device id for a internal LCD*/
        {
            pNvAcpiObject->lcd_handle = dev->handle;
        }
        else if (((device_id >= 0x0111) && (device_id <= 0x0117)) || /* the set
                                        of possible device-ids for a DFP */
                  (device_id == 0x0120) || /* device id for non-LVDS DFP */
                  (device_id == 0x0300))   /* ACPI spec 3.0 specified 
                                              device id for non-LVDS DFP */
        {
            pNvAcpiObject->dvi_handle = dev->handle;
        }
    }

    // arg 0, bits 1:0, 0 = enable events
    control_argument_0.integer.type = ACPI_TYPE_INTEGER;
    control_argument_0.integer.value = 0x0;

    // listify it
    control_argument_list.count = 1;
    control_argument_list.pointer = &control_argument_0;

    // _DOS method takes 1 argument and returns nothing
    status = acpi_evaluate_object(device->handle, "_DOS", &control_argument_list, NULL);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to enable display switch events (%d)!\n", status);
    }

    status = acpi_install_notify_handler(device->handle, ACPI_DEVICE_NOTIFY,
                    nv_acpi_event, pNvAcpiObject);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_add: failed to install event notification handler (%d)!\n", status);
    }
    else
    {
        try_module_get(THIS_MODULE);
        pNvAcpiObject->notify_handler_installed = 1;
    }

    return 0;
}

static int nv_acpi_remove(struct acpi_device *device, int type)
{
    /*
     * This function will cause RM to relinquish control of the VGA ACPI device.
     */
    acpi_status status;
    union acpi_object control_argument_0 = { ACPI_TYPE_INTEGER };
    struct acpi_object_list control_argument_list = { 0, NULL };
    nv_acpi_t *pNvAcpiObject = acpi_driver_data(device);

    // arg 0, bits 1:0, 1 = disable events
    control_argument_0.integer.type = ACPI_TYPE_INTEGER;
    control_argument_0.integer.value = 0x1;

    // listify it
    control_argument_list.count = 1;
    control_argument_list.pointer = &control_argument_0;

    // _DOS method takes 1 argument and returns nothing
    status = acpi_evaluate_object(device->handle, "_DOS", &control_argument_list, NULL);

    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_remove: failed to disable display switch events (%d)!\n", status);
    }

    if (pNvAcpiObject->notify_handler_installed)
    {
        // no status returned for this function
        acpi_os_wait_events_complete(NULL);

        // remove event notifier
        status = acpi_remove_notify_handler(device->handle, ACPI_DEVICE_NOTIFY, nv_acpi_event);
    }

    if (pNvAcpiObject->notify_handler_installed &&
        ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_INFO,
            "NVRM: nv_acpi_remove: failed to remove event notification handler (%d)!\n", status);
    }
    else
    {
        NV_KMEM_CACHE_FREE_STACK(pNvAcpiObject->sp);
        os_free_mem((void *)pNvAcpiObject);
        module_put(THIS_MODULE);
    }

    return status;
}

static void nv_acpi_event(acpi_handle handle, u32 event_type, void *data)
{
    /*
     * This function will handle acpi events from the linux kernel, used
     * to detect notifications from the VGA device.
     */
    nv_acpi_t *pNvAcpiObject = data;
    u32 event_val = 0;
    unsigned long state;
    int status = 0;


    if (event_type == NV_SYSTEM_ACPI_DISPLAY_DEVICE_CYCLE)
    {
        if (pNvAcpiObject->lcd_handle) 
        {
            status = acpi_evaluate_integer(pNvAcpiObject->lcd_handle,
                                           "_DGS", 
                                           NULL, 
                                           &state);
            if (ACPI_FAILURE(status)) 
            {
                nv_printf(NV_DBG_INFO, 
                "NVRM: nv_acpi_event: failed to query _DGS method for LCD\n");
            }
            else if (state)
                event_val |= NV_HOTKEY_STATUS_DISPLAY_ENABLE_LCD;
        }

        if (pNvAcpiObject->crt_handle) 
        {
            status = acpi_evaluate_integer(pNvAcpiObject->crt_handle,
                                           "_DGS",
                                           NULL,
                                           &state);
            if (ACPI_FAILURE(status)) 
            {
                nv_printf(NV_DBG_INFO, 
                "NVRM: nv_acpi_event: failed to query _DGS method for CRT\n");
            }
            else if (state)
                event_val |= NV_HOTKEY_STATUS_DISPLAY_ENABLE_CRT;
        }

        if (pNvAcpiObject->tv_handle) 
        {
            status = acpi_evaluate_integer(pNvAcpiObject->tv_handle,
                                           "_DGS",
                                           NULL,
                                           &state);
            if (ACPI_FAILURE(status)) 
            {
                nv_printf(NV_DBG_INFO,
                "NVRM: nv_acpi_event: failed to query _DGS method for TV\n");
            }
            else if (state)
                event_val |= NV_HOTKEY_STATUS_DISPLAY_ENABLE_TV;
        }

        if (pNvAcpiObject->dvi_handle)
        {
            status = acpi_evaluate_integer(pNvAcpiObject->dvi_handle,
                                           "_DGS",
                                           NULL,
                                           &state);
            if (ACPI_FAILURE(status)) 
            {
                nv_printf(NV_DBG_INFO,
                "NVRM: nv_acpi_event: failed to query _DGS method for DVI\n");
            }
            else if (state)     
                event_val |= NV_HOTKEY_STATUS_DISPLAY_ENABLE_DFP;
        }

        nv_printf(NV_DBG_INFO,
        "NVRM: nv_acpi_event: Event-type 0x%x, Event-val 0x%x\n", 
        event_type, event_val);

        rm_system_event(pNvAcpiObject->sp, event_type, event_val);
    }

    // no unsubscription or re-enable necessary. Once DOD has been set, we are go.
    // once we are subscribed to ACPI events, we don't have to re-subscribe unless
    // unsubscribe.
}

#if defined(NV_ACPI_DEVICE_OPS_HAS_MATCH)
static int nv_acpi_match(struct acpi_device *device, struct acpi_driver *driver)
{
    acpi_handle DOS_method_handler, DOD_method_handler;

    /*
     *
     * attempt to get handles to all of these objects
     * _DOS = enable/disable output switching method
     * _DOD = enumerate all devices attached to the display driver
     * 
     * If handles are available, this is the VGA object.
     *
     */

    if ((!acpi_get_handle(device->handle, "_DOS", &DOS_method_handler)) &&
        (!acpi_get_handle(device->handle, "_DOD", &DOD_method_handler)))
    {
        nv_printf(NV_DBG_INFO, "NVRM: nv_acpi_match: VGA ACPI device detected.\n");

        return 0;
    }
    else
        return (-ENODEV);
}
#endif

/*
 * nvif init function.
 * check if nvif method is present in the acpi namespace.
 * store nvif handle if found. 
 */

RM_STATUS NV_API_CALL nv_acpi_nvif_init(void)
{
    RM_STATUS status = RM_OK;
    
    acpi_walk_namespace(ACPI_TYPE_ANY, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, nv_acpi_find_nvif, NULL, NULL);

    if (!nvif_handle)
        status = RM_ERROR;

    return status;
}
 
acpi_status nv_acpi_find_nvif(
    acpi_handle handle,
    u32 nest_level,
    void *dummy1,
    void **dummy2
)
{
    acpi_handle nvif_method_handler;

    if (!acpi_get_handle(handle, "NVIF", &nvif_method_handler))
    {
        nvif_handle = nvif_method_handler;
    }

    return 0;
}

RM_STATUS NV_API_CALL nv_acpi_nvif_uninit(void)
{
    nvif_handle = NULL;

    return RM_OK;
}

/*
 * executes nvif method.
 */

RM_STATUS NV_API_CALL nv_acpi_nvif_method(
    U032 function, 
    U032 subFunction, 
    void *inParams, 
    U016 inParamSize, 
    U032 *outStatus, 
    void *outData, 
    U016 *outDataSize
)
{
    acpi_status status;
    struct acpi_object_list input;
    struct acpi_buffer output = { ACPI_ALLOCATE_BUFFER, NULL };
    union acpi_object *nvif = NULL;
    union acpi_object nvif_params[3];
    U016 localOutDataSize; 
    U008 localInParams[8];

    if (!nvif_handle)
        return RM_ERROR;

    nvif_params[0].integer.type  =  ACPI_TYPE_INTEGER;
    nvif_params[0].integer.value =  function;

    nvif_params[1].integer.type  =  ACPI_TYPE_INTEGER;
    nvif_params[1].integer.value =  subFunction;

    nvif_params[2].buffer.type  =  ACPI_TYPE_BUFFER;
 
    if (inParams && (inParamSize > 0))
    {
        nvif_params[2].buffer.length  = inParamSize;
        nvif_params[2].buffer.pointer = inParams;
    }
    else
    {
        memset(localInParams, 0, 8);  
        nvif_params[2].buffer.length  = 8;
        nvif_params[2].buffer.pointer = localInParams;
    }

    input.count = 3;
    input.pointer = nvif_params;

    status = acpi_evaluate_object(nvif_handle, NULL, &input, &output);
    if (ACPI_FAILURE(status))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM : nv_acpi_nvif_method : failed to get NVIF data "
            "status %d function %x subFunction %x\n", status, function, subFunction);
        return RM_ERROR;
    }
    
    nvif = output.pointer;
    if (nvif && (nvif->type == ACPI_TYPE_BUFFER) && (nvif->buffer.length >= 4))
    {
        if (outStatus)
        {
            *outStatus = nvif->buffer.pointer[3] << 24 | 
                         nvif->buffer.pointer[2] << 16 |
                         nvif->buffer.pointer[1] << 8  |
                         nvif->buffer.pointer[0];
        }

        if (outData && outDataSize)
        {
            localOutDataSize = nvif->buffer.length - 4;
            if (localOutDataSize <= *outDataSize)
            {
                *outDataSize = NV_MIN(*outDataSize, localOutDataSize);
                memcpy(outData, &nvif->buffer.pointer[4], *outDataSize);
            }
            else
            {
                *outDataSize = localOutDataSize;
                return RM_ERR_BUFFER_TOO_SMALL;
            }
        } 
    }
    else
    {
        nv_printf(NV_DBG_ERRORS, "NVRM : nv_acpi_nvif_method : NVIF data invalid "
            "function %x subFunction %x\n", function, subFunction);
        return RM_ERROR;
    }
 
    return RM_OK; 
}
    
#ifdef DEBUG
#define NV_ACPI_NAME_BUFFER_LENGTH 1024

/*
 * utility function, not normally used except for debugging.
 */
acpi_status nv_acpi_walk_callback(
    acpi_handle handle,
    u32 nest_level,
    void *acpi_name,
    void **dummy2
)
{
    acpi_status status;
    struct acpi_buffer result_buffer;
    char *name = acpi_name;

    result_buffer.length = NV_ACPI_NAME_BUFFER_LENGTH;
    result_buffer.pointer = acpi_name;
    
    status = acpi_get_name(handle, ACPI_SINGLE_NAME, &result_buffer);

    if (ACPI_SUCCESS(status))
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_walk_callback: level: %d, name: %s\n", nest_level, name);
    }

    return 0;
}

/*
 * utility function, not normally used except for debugging.
 */
static void nv_acpi_dump_device(acpi_handle handle)
{
    acpi_status status;
    struct acpi_buffer result_buffer;
    void *acpi_name;

    if (os_alloc_mem(&acpi_name, NV_ACPI_NAME_BUFFER_LENGTH) != RM_OK)
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: failed to allocate ACPI name buffer!\n");
        return;
    }

    result_buffer.length = NV_ACPI_NAME_BUFFER_LENGTH;
    result_buffer.pointer = acpi_name;

    status = acpi_get_name(handle, ACPI_SINGLE_NAME, &result_buffer);

    if (ACPI_SUCCESS(status))
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: Preparing to print namespace for %s\n",
            (char *)acpi_name);
    }
    else
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: nv_acpi_dump_device: Preparing to print namespace for "
            "NAMESPACE INVALID\n");
    }

    acpi_walk_namespace(ACPI_TYPE_ANY, handle, 1, nv_acpi_walk_callback, acpi_name, NULL);

    os_free_mem(acpi_name);
}
#endif // DEBUG

#else // NV_LINUX_ACPI_EVENTS_SUPPORTED

int nv_acpi_init(void)
{
    return 0;
}

int nv_acpi_uninit(void)
{
    return 0;
}

RM_STATUS NV_API_CALL nv_acpi_nvif_init(void)
{
    return RM_ERROR;
}

RM_STATUS NV_API_CALL nv_acpi_nvif_uninit(void)
{
    return RM_ERROR;
}

RM_STATUS NV_API_CALL nv_acpi_nvif_method(
    U032 function,
    U032 subFunction,
    void *inParams,
    U016 inParamSize,
    U032 *outStatus,
    void *outData,
    U016 *outDataSize
)
{
    return RM_ERROR;
}
#endif
